@@ -73,31 +73,19 @@ impl Compiler {
         let start_srcloc = FilePos::new(offset as u32);
         let end_srcloc = FilePos::new((offset + len) as u32);
 
-        let instructions = if let Some(ref mcr) = &context.mach_compile_result {
-            // New-style backend: we have a `MachCompileResult` that will give us `MachSrcLoc` mapping
-            // tuples.
-            collect_address_maps(
-                body_len,
-                mcr.buffer
-                    .get_srclocs_sorted()
-                    .into_iter()
-                    .map(|&MachSrcLoc { start, end, loc }| (loc, start, (end - start))),
-            )
-        } else {
-            // Old-style backend: we need to traverse the instruction/encoding info in the function.
-            let func = &context.func;
-            let mut blocks = func.layout.blocks().collect::<Vec<_>>();
-            blocks.sort_by_key(|block| func.offsets[*block]); // Ensure inst offsets always increase
-
-            let encinfo = self.isa.encoding_info();
-            collect_address_maps(
-                body_len,
-                blocks
-                    .into_iter()
-                    .flat_map(|block| func.inst_offsets(block, &encinfo))
-                    .map(|(offset, inst, size)| (func.srclocs[inst], offset, size)),
-            )
-        };
+        // New-style backend: we have a `MachCompileResult` that will give us `MachSrcLoc` mapping
+        // tuples.
+        let instructions = collect_address_maps(
+            body_len,
+            context
+                .mach_compile_result
+                .as_ref()
+                .unwrap()
+                .buffer
+                .get_srclocs_sorted()
+                .into_iter()
+                .map(|&MachSrcLoc { start, end, loc }| (loc, start, (end - start))),
+        );
 
         FunctionAddressMap {
             instructions: instructions.into(),
@@ -184,30 +172,35 @@ impl wasmtime_environ::Compiler for Compiler {
                 &mut trap_sink,
                 &mut stack_map_sink,
             )
-            .map_err(|error| {
-                CompileError::Codegen(pretty_error(&context.func, Some(isa), error))
-            })?;
+            .map_err(|error| CompileError::Codegen(pretty_error(&context.func, error)))?;
 
-        let unwind_info = context.create_unwind_info(isa).map_err(|error| {
-            CompileError::Codegen(pretty_error(&context.func, Some(isa), error))
-        })?;
+        let unwind_info = context
+            .create_unwind_info(isa)
+            .map_err(|error| CompileError::Codegen(pretty_error(&context.func, error)))?;
 
         let address_transform =
             self.get_function_address_map(&context, &input, code_buf.len() as u32);
 
         let ranges = if tunables.generate_native_debuginfo {
-            let ranges = context.build_value_labels_ranges(isa).map_err(|error| {
-                CompileError::Codegen(pretty_error(&context.func, Some(isa), error))
-            })?;
-            Some(ranges)
+            Some(
+                context
+                    .mach_compile_result
+                    .as_ref()
+                    .unwrap()
+                    .value_labels_ranges
+                    .clone(),
+            )
         } else {
             None
         };
 
+        let timing = cranelift_codegen::timing::take_current();
+        log::debug!("{:?} translated in {:?}", func_index, timing.total());
+        log::trace!("{:?} timing info\n{}", func_index, timing);
+
         let length = u32::try_from(code_buf.len()).unwrap();
         Ok(Box::new(CompiledFunction {
             body: code_buf,
-            jt_offsets: context.func.jt_offsets,
             relocations: reloc_sink.func_relocs,
             value_labels_ranges: ranges.unwrap_or(Default::default()),
             stack_slots: context.func.stack_slots,
@@ -229,7 +222,7 @@ impl wasmtime_environ::Compiler for Compiler {
         types: &TypeTables,
         funcs: PrimaryMap<DefinedFuncIndex, Box<dyn Any + Send>>,
         emit_dwarf: bool,
-        obj: &mut Object,
+        obj: &mut Object<'static>,
     ) -> Result<(PrimaryMap<DefinedFuncIndex, FunctionInfo>, Vec<Trampoline>)> {
         let funcs: crate::CompiledFunctions = funcs
             .into_iter()
@@ -324,7 +317,7 @@ impl wasmtime_environ::Compiler for Compiler {
         &self,
         ty: &WasmFuncType,
         host_fn: usize,
-        obj: &mut Object,
+        obj: &mut Object<'static>,
     ) -> Result<(Trampoline, Trampoline)> {
         let host_to_wasm = self.host_to_wasm_trampoline(ty)?;
         let wasm_to_host = self.wasm_to_host_trampoline(ty, host_fn)?;
@@ -542,17 +535,14 @@ impl Compiler {
                 &mut trap_sink,
                 &mut stack_map_sink,
             )
-            .map_err(|error| {
-                CompileError::Codegen(pretty_error(&context.func, Some(isa), error))
-            })?;
+            .map_err(|error| CompileError::Codegen(pretty_error(&context.func, error)))?;
 
-        let unwind_info = context.create_unwind_info(isa).map_err(|error| {
-            CompileError::Codegen(pretty_error(&context.func, Some(isa), error))
-        })?;
+        let unwind_info = context
+            .create_unwind_info(isa)
+            .map_err(|error| CompileError::Codegen(pretty_error(&context.func, error)))?;
 
         Ok(CompiledFunction {
             body: code_buf,
-            jt_offsets: context.func.jt_offsets,
             unwind_info,
             relocations: reloc_sink.relocs,
             stack_slots: Default::default(),
@@ -657,25 +647,6 @@ impl binemit::RelocSink for RelocSink {
             addend,
         });
     }
-
-    fn reloc_constant(
-        &mut self,
-        _code_offset: binemit::CodeOffset,
-        _reloc: binemit::Reloc,
-        _constant_offset: ir::ConstantOffset,
-    ) {
-        // Do nothing for now: cranelift emits constant data after the function code and also emits
-        // function code with correct relative offsets to the constant data.
-    }
-
-    fn reloc_jt(&mut self, offset: binemit::CodeOffset, reloc: binemit::Reloc, jt: ir::JumpTable) {
-        self.func_relocs.push(Relocation {
-            reloc,
-            reloc_target: RelocationTarget::JumpTable(jt),
-            offset,
-            addend: 0,
-        });
-    }
 }
 
 impl RelocSink {
@@ -787,20 +758,4 @@ impl binemit::RelocSink for TrampolineRelocSink {
             addend,
         });
     }
-    fn reloc_constant(
-        &mut self,
-        _code_offset: binemit::CodeOffset,
-        _reloc: binemit::Reloc,
-        _constant_offset: ir::ConstantOffset,
-    ) {
-        panic!("trampoline compilation should not produce constant relocs");
-    }
-    fn reloc_jt(
-        &mut self,
-        _offset: binemit::CodeOffset,
-        _reloc: binemit::Reloc,
-        _jt: ir::JumpTable,
-    ) {
-        panic!("trampoline compilation should not produce jump table relocs");
-    }
 }
